from dataclasses import dataclass

# Deprecated, soon will be removed
def paisa_to_inr(amount_in_paisa: int):
    return round(amount_in_paisa / 100, ndigits=2)


# Deprecated, soon will be removed
def inr_to_paisa(amount_in_inr: int):
    return int(amount_in_inr * 100)


def to_base_denomination(amount: int):
    return int(amount * 100)


def from_base_denomination(amount: int):
    return round(amount / 100, ndigits=2)


@dataclass(frozen=True)
class Money:
    amount: float
    currency: str



from dataclasses import dataclass
from typing import List
from shapely.geometry import Point, LineString

from geopy.distance import geodesic, Distance
from geo.ellipsoid import distance


@dataclass
class Location:
    lat: float
    lng: float

    def __init__(self, lat: float, lng: float):
        assert -90 <= lat <= 90
        assert -180 <= lng <= 180
        self.lat = lat
        self.lng = lng

    def __eq__(self, o: object) -> bool:
        return (self.lat, self.lng) == (o.lat, o.lng)

    def __hash__(self) -> int:
        return hash((self.lat, self.lng))

    def as_point(self):
        return Point(self.lng, self.lat)

    def distance_to(self, other) -> Distance:
        loc1 = (self.lat, self.lng)
        loc2 = (other.lat, other.lng)
        return geodesic(loc1, loc2)

    def distance_to_fast(self, other) -> Distance:
        loc1 = (self.lat, self.lng)
        loc2 = (other.lat, other.lng)
        return Distance(kilometers=distance(loc1, loc2) / 1000)


@dataclass
class Path:
    def __init__(self, locations: List[Location]):
        assert len(locations) >= 2
        points = [location.as_point() for location in locations]
        self._path = LineString([[p.x, p.y] for p in points])

    def as_list(self):
        return [Location(lat=y, lng=x) for x, y in self._path.coords]



from shuttlis.money import paisa_to_inr, inr_to_paisa
from shuttlis.money import to_base_denomination, from_base_denomination


def test_paisa_to_inr():
    assert paisa_to_inr(2097) == 20.97
    assert paisa_to_inr(208) == 2.08
    assert paisa_to_inr(2090) == 20.9


def test_inr_to_paisa():
    assert inr_to_paisa(20.97) == 2097
    assert inr_to_paisa(20.9) == 2090
    assert inr_to_paisa(2.08) == 208


def test_from_base_denomination():
    assert from_base_denomination(2097) == 20.97
    assert from_base_denomination(208) == 2.08
    assert from_base_denomination(2090) == 20.9


def test_to_base_denomination():
    assert to_base_denomination(20.97) == 2097
    assert to_base_denomination(20.9) == 2090
    assert to_base_denomination(2.08) == 208



from dataclasses import dataclass
from datetime import datetime
from typing import Dict, Optional
from uuid import UUID

from shuttlis.time import from_iso_format


@dataclass
class After:
    """
    This class represents how we make the cursor for cursor based pagination.
    The default logic is to combine the created_at and uuid of the object to create
    a cursor which is then used by the frontend. This class is used by Cursor and
    the Paginator class.

    The Cursor and Paginator classes are configurable to use any other class. Just make
    sure that the class has the `from_string`, `from_data` and `__str__` implemented.
    """

    id: UUID
    date: datetime

    @classmethod
    def from_string(cls, after) -> "After":
        offset = after.split("|")
        dt = from_iso_format(offset[0])
        id = UUID(offset[1])
        return cls(id, dt)

    @classmethod
    def from_data(cls, data) -> "After":
        return cls(data.id, data.created_at)

    def __str__(self):
        return f"{self.date.isoformat()}|{self.id}"


@dataclass
class Cursor:
    """
    This represents the cursor that the frontend will send to the backend. Consider
    the following requests

    /rs?limit=10
    /rs?after=2019-01-08T16:59:48+00:00|ce581a45-48a7-4ff8-b627-9ff2d43d2ea6
    /rs?after=2019-01-08T16:59:48+00:00|ce581a45-48a7-4ff8-b627-9ff2d43d2ea6&limit=9

    One can directly construct the Cursor object using the query params obj
    e.g. Flask
    >>> from flask import request
    >>> c = Cursor.from_dict(request.args)

    or if you have extracted out the params yourself then you can use the from_strings
    classmethod
    >>> c = Cursor.from_strings(
    >>>     "2019-01-08T16:59:48.377325+00:00|ce581a45-48a7-4ff8-b627-9ff2d43d2ea6",
    >>>     "10",
    >>> )

    Then one can use this object anywhere to do pagination at let's say db level
    >>> from sqlalchemy import or_, and_
    >>> from store import db, MyModel
    >>> query = db.session.query(MyModel)
    >>> if c.after:
    >>>     query = query.filter(
    >>>         or_(
    >>>             MyModel.created_at > c.after_date,
    >>>             and_(MyModel.created_at == c.after_date, MyModel.id > c.after_id),
    >>>         )
    >>>     )
    >>> return query.order_by(MyModel.created_at, MyModel.id).limit(c.limit).all()
    """

    after: Optional[After]
    limit: Optional[int] = 10

    @property
    def after_id(self) -> Optional[UUID]:
        return self.after.id if self.after else None

    @property
    def after_date(self) -> Optional[datetime]:
        return self.after.date if self.after else None

    def as_dict(self):
        dikt = {"limit": self.limit}

        if self.after:
            dikt["after"] = str(self.after)

        return dikt

    @classmethod
    def from_dict(cls, json, custom_class=After) -> "Cursor":
        after = None

        if json.get("after"):
            after = custom_class.from_string(json["after"])

        return Cursor(after, json.get("limit", 10))

    @classmethod
    def from_strings(cls, after: str = None, limit: str = "10") -> "Cursor":
        if after:
            after = After.from_string(after)

        if limit is None:
            limit = 10

        return Cursor(after, int(limit))


@dataclass
class Paginator:
    """
    This represents the pagination info that the backend will send to the frontend.
    Consider the following response.
    >>> {
    >>>     "data": [ ... ],
    >>>     "meta": {
    >>>         "cursor": {
    >>>             "limit": 10,
    >>>             "total": 100,
    >>>             "last": "2019-01-08T16:59:48|ce581a45-48a7-4ff8-b627-9ff2d43d2ea6"
    >>>         }
    >>>     }
    >>> }

    Now the frontend can use this info to query for the next page.
    Now let's look at how this can be constructed.

    Basic idea is that the objects that are involved in pagination should have `id` and
    `created_at` fields that are used to construct this data.

    This is how one can use this with flask
    >>> import json
    >>> from flask import Response

    Somewhere in your controller
    >>> data = [ ... ] # get from somewhere
    >>> p = Paginator.from_data(data)
    >>> payload = {"data": data, "meta": {"cursor": p.as_dict()}}
    >>> return Response(json.dumps(payload), mimetype="application/json")
    """

    after: Optional[After]
    limit: int = 10
    total: Optional[int] = None

    @property
    def after_id(self) -> Optional[UUID]:
        return self.after.id if self.after else None

    @property
    def after_date(self) -> Optional[datetime]:
        return self.after.date if self.after else None

    @classmethod
    def from_data(cls, data, page_size=10, total=None, custom_cls=After) -> "Paginator":
        cur = None

        if data:
            last = data[-1]
            cur = custom_cls.from_data(last)

        return cls(after=cur, limit=page_size, total=total)

    def as_dict(self) -> Dict:
        json_res = {"page_size": self.limit}

        if self.total:
            json_res["total"] = self.total

        if self.after:
            json_res["last"] = str(self.after)

        return json_res

    @classmethod
    def from_dict(cls, dikt, custom_cls=After) -> "Paginator":
        after = None

        if dikt.get("last"):
            after = custom_cls.from_string(dikt.get("last"))

        return cls(after, dikt.get("limit", 10), dikt.get("total"))



import json
import pytest
import logging

from shuttlis.log import configure_logging


def test_log_format_defaults_to_json(capsys):
    logger = configure_logging("pyshuttlis", "DEBUG")
    logger.debug("hey.there")

    captured = capsys.readouterr().err
    assert json.loads(captured)


def test_log_format_can_be_set_to_console(capsys):
    logger = configure_logging("pyshuttlis", "DEBUG", log_format="console")
    logger.debug("hey.there")

    captured = capsys.readouterr().err
    with pytest.raises(json.decoder.JSONDecodeError):
        json.loads(captured)

    assert "hey.there" in captured


def test_extra_is_printed_in_json(capsys):
    logger = configure_logging("pyshuttlis", "DEBUG", log_format="json")
    logger.debug("hey.there", extra={"one": "two"})

    captured = capsys.readouterr().err
    assert "one" in captured
    assert "two" in captured


def test_extra_is_printed_in_console(capsys):
    logger = configure_logging("pyshuttlis", "DEBUG", log_format="console")
    logger.debug("hey.there", extra={"one": "two"})

    captured = capsys.readouterr().err
    assert "one" in captured
    assert "two" in captured


@pytest.mark.xfail
def test_prints_out_exc(capsys):
    """
    TODO: This doesn't seem to work, ValueError isn't logged!
    """
    logger = configure_logging("pyshuttlis", "DEBUG")
    try:
        raise ValueError
    except ValueError:
        logger.error("an.error", exc_info=True)

    captured = capsys.readouterr().err
    assert "an.error" in captured
    assert "ValueError" in captured


@pytest.fixture(autouse=True)
def reset_root_logger():
    yield

    # Pytest's mocks cause issues unless we reset this
    logging.getLogger().handlers = []



from .military_time import *  # noqa: F403, F401
from .weekday import *  # noqa: F403, F401
from .time import (
    TimeWindow,
    TimeDeltaWindow,
    time_now,
    from_iso_format,
    date_from_iso_format,
)  # noqa: F401



import pytest

from shuttlis.utils import one_or_none, one, id_map, next_or_none, pairwise


@pytest.mark.parametrize(
    "lizt,key,expected",
    [
        ([1, 2, 3], lambda x: x == 2, 2),
        (["a", "b", "c"], lambda x: x == "c", "c"),
        (["a", "b", "c"], lambda x: x == "d", None),
    ],
)
def test_one_or_none(lizt, key, expected):
    actual = one_or_none(lizt, key)
    assert expected == actual


def test_one_or_none_raises_assertion_error_if_more_than_one_match():
    with pytest.raises(AssertionError):
        one_or_none([1, 2, 2, 2], lambda x: x == 2)


@pytest.mark.parametrize(
    "lizt,key,expected",
    [
        ([1, 2, 3], lambda x: x == 2, 2),
        (["a", "b", "c"], lambda x: x == "c", "c"),
        (["a", "b", "c", "c"], lambda x: x == "c", "c"),
        (["a", "b", "c"], lambda x: x == "d", None),
    ],
)
def test_next_or_none(lizt, key, expected):
    actual = next_or_none(lizt, key)
    assert expected == actual


@pytest.mark.parametrize(
    "lizt,key,expected",
    [([1, 2, 3], lambda x: x == 2, 2), (["a", "b", "c"], lambda x: x == "c", "c")],
)
def test_one(lizt, key, expected):
    actual = one(lizt, key)
    assert expected == actual


def test_one_raises_assertion_error_if_more_than_one_match():
    with pytest.raises(AssertionError):
        one([1, 2, 2, 2], lambda x: x == 2)


def test_one_raises_assertion_error_if_none_match():
    with pytest.raises(AssertionError):
        one([1, 2, 2, 2], lambda x: x == 4)


@pytest.mark.parametrize(
    "lizt,key,expected", [(["a", "ab", "abc"], len, {1: "a", 2: "ab", 3: "abc"})]
)
def test_id_map(lizt, key, expected):
    assert expected == id_map(lizt, key=key)


def test_id_map_with_multiple_items_raises_error():
    with pytest.raises(AssertionError):
        id_map(["a", "b", "c"], key=len)


def test_pairwise():
    assert pairwise([1, 2, 3]) == ([(1, 2), (2, 3)])
    assert pairwise([]) == []



from dataclasses import dataclass
from datetime import datetime, time, date
from functools import total_ordering

tz_datetime = datetime


class TimezoneNotFound(Exception):
    pass


class TimezoneMismatch(Exception):
    pass


@dataclass(frozen=True)
class TimeDelta:
    hr: int = 0
    min: int = 0


@total_ordering
class MilitaryTime:
    def __init__(self, time, tzinfo):
        hr, min = divmod(time, 100)
        assert 0 <= min < 60
        assert 0 <= hr < 24

        self.hr = hr
        self.min = min
        self.tz = tzinfo

    @classmethod
    def from_hr_min(cls, hr: int, min: int, tzinfo):
        return cls(hr * 100 + min, tzinfo)

    @classmethod
    def extract_from_datetime(cls, dt: datetime):
        if not dt.tzinfo:
            raise TimezoneNotFound("No timezone present in datatime object")

        return cls.from_hr_min(dt.hour, dt.minute, dt.tzinfo)

    @property
    def time(self):
        return self.hr * 100 + self.min

    def to_time(self, tzinfo) -> time:
        return time(hour=self.hr, minute=self.min, tzinfo=tzinfo)

    def tz_unaware_time(self) -> time:
        return time(hour=self.hr, minute=self.min)

    def combine(self, dt: date) -> datetime:
        return self.tz.localize(datetime.combine(dt, self.tz_unaware_time()))

    def __eq__(self, other):
        return (self.hr, self.min, self.tz.zone) == (other.hr, other.min, other.tz.zone)

    def __str__(self):
        return f"{self.hr:02}{self.min:02}:{self.tz.zone}"

    def __repr__(self):
        return f"<MilitaryTime: {str(self)}>"

    def __add__(self, other: TimeDelta):
        min = (self.min + other.min) % 60
        additional_hrs = (self.min + other.min) // 60
        hr = (self.hr + other.hr + additional_hrs) % 24
        return MilitaryTime.from_hr_min(hr, min, self.tz)

    def __sub__(self, other: TimeDelta):
        min = (self.min - other.min) % 60
        additional_hrs = (self.min - other.min) // 60
        hr = (self.hr - other.hr + additional_hrs) % 24
        return MilitaryTime.from_hr_min(hr, min, self.tz)

    def __hash__(self):
        return hash((self.hr, self.min))

    def __lt__(self, other):
        if self.tz.zone != other.tz.zone:
            raise TimezoneMismatch(
                f"Recieved different timezones as {self.tz.zone} and {other.tz.zone}"
            )

        return (self.hr * 100, self.min) < (other.hr * 100, other.min)



from dataclasses import dataclass
from datetime import datetime, timedelta, date
from typing import Optional, Generator

import pytz


def time_now() -> datetime:
    return pytz.utc.localize(datetime.utcnow())


def from_iso_format(d: str, tz=pytz.utc) -> datetime:
    return datetime.fromisoformat(d).astimezone(tz)


def maybe_datetime_from_iso_format(d: str, tz=pytz.utc) -> Optional[datetime]:
    return datetime.fromisoformat(d).astimezone(tz) if d else None


def date_from_iso_format(d: str, tz=pytz.utc) -> date:
    time = datetime.fromisoformat(d).astimezone(tz) + timedelta(hours=12)
    return time.date()


def maybe_date_from_iso_format(d: str, tz=pytz.utc) -> Optional[date]:
    time = datetime.fromisoformat(d).astimezone(tz) + timedelta(hours=12) if d else None
    return time.date() if time else None


@dataclass(frozen=True)
class TimeDeltaWindow:
    lower: Optional[timedelta]
    upper: Optional[timedelta]

    @classmethod
    def from_minutes(cls, lower: int, upper: int):
        lower = timedelta(minutes=lower)
        upper = timedelta(minutes=upper)
        return cls(lower, upper)

    @classmethod
    def from_days(cls, lower: int, upper: int):
        lower = timedelta(days=lower)
        upper = timedelta(days=upper)
        return cls(lower, upper)


class TimeWindowError(Exception):
    pass


@dataclass
class TimeWindow:
    def __init__(self, from_date: datetime = None, to_date: datetime = None):
        from_date = from_date or pytz.utc.localize(datetime.min)
        to_date = to_date or pytz.utc.localize(datetime.max)

        assert to_date >= from_date

        self.__from_date = from_date
        self.__to_date = to_date

    @property
    def from_date(self):
        return self.__from_date

    @property
    def to_date(self):
        return self.__to_date

    def __contains__(self, item):
        return self.from_date <= item <= self.to_date

    def __str__(self):
        return f"{self.from_date.isoformat()} - {self.to_date.isoformat()}"

    def __hash__(self):
        return hash((self.from_date, self.to_date))

    def intersects(self, other: "TimeWindow"):
        c1 = self.from_date in other or self.to_date in other
        c2 = other.from_date in self or other.to_date in self
        return c1 or c2

    def intersection(self, other: "TimeWindow"):
        if not self.intersects(other):
            raise ValueError("No intersection possible")

        return TimeWindow(
            from_date=max(self.from_date, other.from_date),
            to_date=min(self.to_date, other.to_date),
        )

    def union(self, other: "TimeWindow"):
        if not self.intersects(other):
            raise ValueError("No union possible")

        return TimeWindow(
            from_date=min(self.from_date, other.from_date),
            to_date=max(self.to_date, other.to_date),
        )

    @classmethod
    def around(cls, dt: datetime, td_window: TimeDeltaWindow) -> "TimeWindow":
        fr, to = None, None

        if td_window.lower is not None:
            fr = dt - td_window.lower

        if td_window.upper is not None:
            to = dt + td_window.upper

        return cls(fr, to)

    @classmethod
    def around_now(cls, td: timedelta) -> "TimeWindow":
        td_window = TimeDeltaWindow(td, td)
        return cls.around(time_now(), td_window)

    def dates(self) -> Generator[datetime, None, None]:
        return self.range(timedelta(days=1))

    def range(self, td: timedelta) -> Generator[datetime, None, None]:
        temp_date = self.from_date

        while temp_date <= self.to_date:
            yield temp_date
            temp_date += td



import logging


class AddContextualFieldFilter(logging.Filter):
    def __init__(self, field_name: str, field_value: str) -> None:
        self._field_name = field_name
        self._field_value = field_value

    def filter(self, record: logging.LogRecord) -> bool:
        record.__dict__[self._field_name] = self._field_value
        return True


class MaskPasswordsInLogs(logging.Filter):
    def filter(self, record: logging.LogRecord) -> bool:
        if record.__dict__.get("password") is not None:
            record.__dict__["password"] = "******"
        return True



from uuid import uuid4
from datetime import datetime, date

from shuttlis.serialization import serialize
from shuttlis.time import MilitaryTime
from pytz import timezone

IST_TIMEZONE = timezone("Asia/Kolkata")


def test_serialize_with_None():
    res = serialize(None)
    assert res is None


def test_serialize_with_military_time():
    time = MilitaryTime(800, IST_TIMEZONE)
    res = serialize(time)
    assert {"time": 800, "timezone": "Asia/Kolkata"} == res


def test_serialize_with_uuid():
    uuid = uuid4()
    res = serialize(uuid)
    assert str(uuid) == res


def test_serialize_with_datetime():
    ts = datetime(2019, 1, 1, 8, 30, 0)
    res = serialize(ts)
    assert "2019-01-01T08:30:00" == res
    assert datetime.fromisoformat(res) == ts


def test_serialize_with_date():
    dt = date(2019, 1, 1)
    res = serialize(dt)
    assert "2019-01-01" == res
    assert date.fromisoformat(res) == dt


def test_serialize_with_byte_array():
    bytes = b"Hey Dude!!!!"
    res = serialize(bytes)
    assert "Hey Dude!!!!" == res


def test_serialize_with_list():
    uuid = uuid4()
    res = serialize([uuid])
    assert [str(uuid)] == res


def test_serialize_with_frozen_set_converts_it_to_list():
    uuid = uuid4()
    res = serialize(frozenset([uuid]))
    assert [str(uuid)] == res


def test_serialize_with_set_converts_it_to_list():
    uuid = uuid4()
    res = serialize({uuid})
    assert [str(uuid)] == res



import pytest

from shuttlis.geography import Location


@pytest.mark.parametrize("lat,lng", [(90.0, 180.0), (-90.0, -180.0)])
def test_location_constructer_with_valid_values(lat, lng):
    loc = Location(lat, lng)
    assert lat == loc.lat
    assert lng == loc.lng


@pytest.mark.parametrize("lat,lng", [(90.0, 180.0), (-90.0, -180.0)])
def test_location_equality_hash(lat, lng):
    loc = Location(lat, lng)
    loc2 = Location(lat, lng)
    assert loc == loc2
    assert hash(loc) == hash(loc2)


@pytest.mark.skip("Version Mismatch")
def test_location_distance_to():
    l1 = Location(28.434816, 77.049327)
    l2 = Location(28.443195, 77.057225)

    assert 1208.6872544594537 == l1.distance_to(l2).meters


def test_location_distance_to_fast():
    l1 = Location(28.434816, 77.049327)
    l2 = Location(28.443195, 77.057225)

    assert 906.2925481254175 == l1.distance_to_fast(l2).meters


@pytest.mark.parametrize("lat,lng", [(90, 190), (90.1, 179), (90.1, 180.001)])
def test_location_constructer_disallows_lat_lng_values_which_are_out_of_range(lat, lng):
    with pytest.raises(AssertionError):
        Location(lat, lng)



import json
import logging

from .utils import _sanitize_stacktrace_for_json_fields


class ConsoleFormatter(logging.Formatter):
    def format(self, record: logging.LogRecord) -> str:
        level = record.levelname
        timestamp = self.formatTime(record, "%H:%M:%S")
        message = record.getMessage()
        extra_fields = self._extract_extra_fields(record)
        extra_txt = " ".join([f"{k}={v}" for k, v in extra_fields.items()])
        funcname = f"{record.module}.{record.funcName}"

        return f"[{level}] -- {timestamp} :: (#{funcname}) :: {message} [{extra_txt}]"

    def _extract_extra_fields(self, record: logging.LogRecord) -> dict:
        dummy_record = logging.LogRecord(*["dummy"] * 7)
        default_fields = dummy_record.__dict__.keys()
        return {k: v for k, v in record.__dict__.items() if k not in default_fields}


class JSONFormatter(logging.Formatter):
    def format(self, record: logging.LogRecord) -> str:
        json_dict = {}

        if record.exc_info:
            json_dict["stack_trace"] = _sanitize_stacktrace_for_json_fields(
                *record.exc_info
            )

            # mark exception info as none to consume the exception here and stop propogation
            # where it might get logged to sys.stderr.
            # If we need to reraise it we can log it as an exception on the root logger
            record.exc_info = None

        json_dict.update(record.__dict__)

        # Already replaced this with 'stack_trace', no longer needed
        assert json_dict["exc_info"] is None
        del json_dict["exc_info"]

        # We don't need these 2 fields as message field has msg % args now
        json_dict["message"] = record.getMessage()
        del json_dict["msg"]
        del json_dict["args"]

        json_dict["level"] = json_dict["levelname"]
        del json_dict["levelname"]

        try:
            return json.dumps(json_dict)
        except (TypeError, OverflowError):
            return json.dumps(
                {"message": "An unrecoverable exception occured while logging"}
            )



from shuttlis.set import KeyedSet


class Foo(object):
    def __init__(self, name):
        self.name = name

    def __repr__(self):
        return "%s(%r)" % (type(self).__name__, self.name)


def getname(o):
    return o.name


def test_keyed_set_with_overlapping_values():
    objs = [Foo("Joe"), Foo("Jim"), Foo("Tom"), Foo("Jim")]

    s = KeyedSet(objs, key=getname)
    assert s == KeyedSet([Foo("Jim"), Foo("Joe"), Foo("Tom")], key=getname)


def test_adding_same_items_to_keyed_set():
    s1 = KeyedSet([Foo("Jim"), Foo("Joe"), Foo("Tom")], key=getname)
    n_set = KeyedSet([Foo("Jim"), Foo("Joe"), Foo("Tom")], key=getname)
    s1.add(Foo("Joe"))

    assert s1 == n_set


def test_membership_in_keyed_set():
    s1 = KeyedSet([Foo("Jim"), Foo("Joe"), Foo("Tom")], key=getname)

    assert Foo("Jim") in s1


def test_membershipt_with_key_returns_false():
    s1 = KeyedSet([Foo("Jim"), Foo("Joe"), Foo("Tom")], key=getname)

    assert "Jim" not in s1


def test_union_with_normal_set():
    s1 = KeyedSet([Foo("Jim"), Foo("Joe"), Foo("Tom")], key=getname)
    s2 = {Foo("Luc"), Foo("Jim"), Foo("Dan")}

    expected = KeyedSet([Foo("Dan"), Foo("Jim"), Foo("Luc"), Foo("Joe"), Foo("Tom")])

    assert expected == s1 | s2


def test_intersection_with_other_keyed_set():
    s1 = KeyedSet([Foo("Jim"), Foo("Joe"), Foo("Tom")], key=getname)
    s2 = {Foo("Luc"), Foo("Jim"), Foo("Dan")}

    assert KeyedSet([Foo("Jim")]) == s1 & s2


def test_asymtric_diff_with_keyed_set():
    s1 = KeyedSet([Foo("Jim"), Foo("Joe"), Foo("Tom")], key=getname)
    s2 = KeyedSet([Foo("Luc"), Foo("Jim"), Foo("Dan")], key=getname)

    assert KeyedSet([Foo("Joe"), Foo("Tom")]) == s1 - s2
    assert KeyedSet([Foo("Luc"), Foo("Dan")]) == s2 - s1



from unittest.mock import MagicMock
from uuid import uuid4

import pytest

from shuttlis.pagination import After, Cursor, Paginator
from shuttlis.time import time_now, dataclass


@pytest.mark.parametrize("id,date", [(uuid4(), time_now())])
def test_from_string_and_str_are_dual(id, date):
    after = After(id, date)
    assert after == After.from_string(str(after))


def test_construction_of_cursor():
    after = After(uuid4(), time_now())
    json_req = {"after": str(after)}
    assert Cursor(after, 10) == Cursor.from_dict(json_req)


def test_as_dict_of_cursor():
    cursor = Cursor(after=None)

    assert {"limit": 10} == cursor.as_dict()


def test_construction_of_cursor_from_string():
    after = After(uuid4(), time_now())
    string = str(after)

    assert Cursor(after, 10) == Cursor.from_strings(string, None)


def test_construction_of_paginator():
    obj = MagicMock(id=uuid4(), created_at=time_now())
    paginator = Paginator.from_data([obj])

    assert str(After(obj.id, obj.created_at)) == paginator.as_dict()["last"]


def test_construction_of_paginator_from_dict():
    dikt = {
        "limit": 10,
        "total": 100,
        "last": "2019-01-08T16:59:48|ce581a45-48a7-4ff8-b627-9ff2d43d2ea6",
    }
    after = After.from_string(dikt["last"])
    paginator = Paginator(after, limit=10, total=100)

    assert paginator == Paginator.from_dict(dikt)


def test_as_json_for_custom_paginator_and_paginator_are_same_for_similar_objs():
    id, date = uuid4(), time_now()

    p_obj = MagicMock(id=id, created_at=date)
    cp_obj = MagicMock(id=id, destroyed_at=date)

    @dataclass
    class CustomAfter(After):
        @classmethod
        def from_data(cls, data) -> "CustomAfter":
            return cls(data.id, data.destroyed_at)

    cp = Paginator.from_data([cp_obj], custom_cls=CustomAfter)
    p = Paginator.from_data([p_obj])

    assert p.as_dict() == cp.as_dict()



from uuid import uuid4

import pytest
from voluptuous import Invalid, MultipleInvalid
from datetime import date

from shuttlis.validators import (
    uuid_validator,
    csv_uuids,
    location,
    military_time,
    coerce_timezone,
)
from shuttlis.time import time_now
from shuttlis.validators import datetime_validator, date_validator


def test_datetime_validator_converts_iso_string_to_datetime():
    time = time_now()
    assert time == datetime_validator(time.isoformat())


@pytest.mark.parametrize(
    "invalid_datetime", ["2013-02-01T12:12:12+5:30", 123123, "random"]
)
def test_datetime_validator_throws_error_when_value_is_not_uuid(invalid_datetime):
    with pytest.raises(ValueError) as e:
        datetime_validator(invalid_datetime)


def test_date_validator_converts_iso_string_to_date():
    assert date(2019, 6, 30) == date_validator("2019-06-30")


@pytest.mark.parametrize("invalid_date", ["02-30-2019", 123123, "random"])
def test_date_validator_throws_error_when_value_is_not_iso_format(invalid_date):
    with pytest.raises(ValueError) as e:
        date_validator(invalid_date)


def test_uuid_validator_converts_string_to_uuid():
    r_uuid = str(uuid4())
    assert r_uuid == uuid_validator(r_uuid)


@pytest.mark.parametrize("invalid_uuid", ["abc", 123])
def test_uuid_validator_throws_error_when_value_is_not_uuid(invalid_uuid):
    with pytest.raises(Invalid) as e:
        uuid_validator(invalid_uuid)


def test_csv_uuid_validator():
    random_uuids = [str(uuid4()) for _ in range(5)]
    csv = ",".join(random_uuids)
    assert random_uuids == csv_uuids(csv)


@pytest.mark.parametrize(
    "lat,lng", [(90.0, 180.0), (-90.0, -180.0), (23.234, 23.234234)]
)
def test_location_validator(lat, lng):
    loc = {"lat": lat, "lng": lng}
    assert loc == location(loc)


@pytest.mark.parametrize("lat,lng", [(90, 190), (90.1, 179), (90.1, 180.001)])
def test_location_validator_disallows_lat_lng_values_which_are_out_of_range(lat, lng):
    with pytest.raises(MultipleInvalid):
        location({"lat": lat, "lng": lng})


@pytest.mark.parametrize("time", [1239, 1903, 209, 905, 945, 0, 2359])
def test_military_time_validator(time):
    assert time == military_time(time)


@pytest.mark.parametrize("time", [20983, -1, 2400, 1760, 161, 2546])
def test_military_validator_disallows_time_values_which_are_out_of_range(time):
    with pytest.raises(Invalid):
        military_time(time)


@pytest.mark.parametrize("tz", ["Asia/Kolkata", "Africa/Casablanca"])
def test_timezone_validator(tz):
    assert tz == coerce_timezone(tz).zone


@pytest.mark.parametrize("tz", ["Asia", "alksjdf"])
def test_timezone_validator_fail(tz):
    with pytest.raises(Invalid):
        coerce_timezone(tz)



import uuid
from uuid import UUID

import pytz as pytz
from voluptuous import Invalid, Optional, Schema, All, Any, Required, Range

from shuttlis.pagination import After
from shuttlis.time import from_iso_format, date_from_iso_format


def valid_after(after: str) -> str:
    try:
        After.from_string(after)
        return after
    except:
        raise Invalid("after")


def csv_strings(value):
    try:
        return [string for string in value.split(",")]
    except Exception as e:
        raise Invalid(str(e))


def military_time(time: int) -> int:
    hr, min = divmod(time, 100)
    if not (0 <= hr < 24 and 0 <= min < 60):
        raise Invalid("Invalid Military Time")
    return time


def coerce_timezone(tz):
    try:
        return pytz.timezone(tz)
    except Exception as e:
        raise Invalid("Invalid timezone")


military_time_tz = Schema(
    {Required("time"): All(int, military_time), Required("timezone"): coerce_timezone}
)


location = Schema(
    {
        "lat": All(Any(float, int), Range(min=-90, max=90)),
        "lng": All(Any(float, int), Range(min=-180, max=180)),
    }
)


def int_validator(value: str) -> str:
    try:
        int(value)
        return value
    except Exception as e:
        raise Invalid(value)


pagination = Schema(
    {Optional("limit"): int_validator, Optional("after"): All(str, valid_after)}
)


def csv_uuids(string) -> [UUID]:
    return [uuid_validator(s) for s in string.split(",")]


def uuid_validator(value: str) -> str:
    try:
        uuid.UUID(value)
        return value
    except Exception as e:
        raise Invalid(value)


def datetime_validator(value):
    try:
        return from_iso_format(value)
    except Exception as e:
        raise ValueError(str(e))


def date_validator(value):
    try:
        return date_from_iso_format(value)
    except Exception as e:
        raise ValueError(str(e))



from datetime import datetime, date
from enum import Enum
from uuid import UUID
from shuttlis.time import MilitaryTime

primitive = (int, float, str, bool)


def is_primitive(thing):
    return isinstance(thing, primitive)


def serialize(obj, decoder="utf8"):
    if obj is None:
        return None
    if is_primitive(obj):
        return obj
    if isinstance(obj, Enum):
        return obj.value
    if isinstance(obj, datetime):
        return obj.isoformat()
    if isinstance(obj, MilitaryTime):
        return {"time": obj.time, "timezone": str(obj.tz)}
    if isinstance(obj, date):
        return obj.isoformat()
    if isinstance(obj, UUID):
        return str(obj)
    if isinstance(obj, bytes):
        return str(obj.decode(decoder))
    if isinstance(obj, list):
        return [serialize(v) for v in obj]
    if isinstance(obj, set):
        return [serialize(v) for v in obj]
    if isinstance(obj, frozenset):
        return [serialize(v) for v in obj]
    if isinstance(obj, dict):
        return {k: serialize(v) for k, v in obj.items()}

    return {k: serialize(v) for k, v in obj.__dict__.items()}



from setuptools import setup, find_packages

setup(
    name="pyshuttlis",
    version="0.1.46",
    description="Utilities",
    url="https://github.com/shuttl-tech/pyshuttlis",
    author="Shuttl",
    author_email="sherub.thakur@shuttl.com",
    license="MIT",
    packages=find_packages(),
    classifiers=["Programming Language :: Python :: 3.7"],
    install_requires=["shapely", "pytz", "voluptuous", "geopy", "geo-py", "cython"],
    extras_require={
        "test": ["pytest", "pytest-runner", "pytest-cov", "pytest-pep8"],
        "dev": ["flake8"],
    },
)



# Adding this directly here as I was unable to find a package for this.
# from http://code.activestate.com/recipes/576932-sets-with-a-custom-equalityuniqueness-function/
"""
FrozenKeyedSet and KeyedSet are set implementations that use
a custom element equality function, so that set membership and
uniqueness is not based on the element's __eq__ but on a user-supplied
function.

>>> class Foo(object):
...   def __init__(self, name): self.name = name
...   def __repr__(self): return "%s(%r)" % (type(self).__name__, self.name)
>>> def getname(o):
...   return o.name
>>> objs = [Foo('Joe'), Foo('Jim'), Foo('Tom'), Foo('Jim')]
>>> s = KeyedSet(objs, key=getname)
>>> s
KeyedSet([Foo('Jim'), Foo('Joe'), Foo('Tom')])
>>> s.add(Foo('Joe'))
>>> len(s)
3
>>> joe = Foo('Joe')
>>> joe in s
True
>>> 'Joe' in s
False
>>> s2 = set([Foo('Luc'), Foo('Jim'), Foo('Dan')])
>>> s | s2
KeyedSet([Foo('Dan'), Foo('Jim'), Foo('Luc'), Foo('Joe'), Foo('Tom')])
>>> s & s2
KeyedSet([Foo('Jim')])
>>> s2f = FrozenKeyedSet(s2, key=getname)
>>> s2f - s
FrozenKeyedSet([Foo('Dan'), Foo('Luc')])
>>> for elem in s ^ s2f:
...   print(elem.name)
Dan
Luc
Joe
Tom
>>> s.copy()
KeyedSet([Foo('Jim'), Foo('Joe'), Foo('Tom')])
>>> # FrozenKeyedSet can be used as dictionary key
... d = {}
>>> d[s2f] = ['anything', 'else']

Copyright (C) 2009 Gabriel A. Genellina

"""

__author__ = "Gabriel A. Genellina"
__version__ = "$Revision: 1.12 $"[11:-2]

from collections.abc import Set, MutableSet


# Abstract classes Set and MutableSet define only the operators (&, &=, etc.)
# but not the corresponding "worded" methods (intersection, intersection_update, etc.)
# The former require a Set second argument, the later accept any iterable.
# This helper function is used to build the "worded" method variant on top of
# the corresponding operator.
def _build_variant(cls, opname):
    fn = getattr(cls, opname)
    if hasattr(fn, "im_func"):
        fn = fn.im_func

    def method(self, other, fn=fn):
        if not isinstance(other, Set):
            other = self._from_iterable(other)
        return fn(self, other)

    return method


class FrozenKeyedSet(Set):
    """A frozen set that uses a custom element equality function."""

    # "named" methods like those of frozenset, in addition to operators
    intersection = _build_variant(Set, "__and__")
    union = _build_variant(Set, "__or__")
    difference = _build_variant(Set, "__sub__")
    symmetric_difference = _build_variant(Set, "__xor__")
    issubset = _build_variant(Set, "__le__")
    issuperset = _build_variant(Set, "__ge__")

    def __init__(self, iterable, key=lambda x: x):
        """Create a FrozenKeyedSet from iterable; key function determines uniqueness.

        `key` must be a callable taking a single argument; it is
        applied to every item in the iterable, and its result
        is used to determine set membership. That is, if key(item)
        returns the same for two items, only one of them will be
        in the set.
        The key *must* return a hashable object.
        """
        self._items = dict((key(item), item) for item in iterable)
        self._key = key

    # Implementation of abstract methods from the Set ABC

    def __iter__(self):
        return iter(self._items.values())

    def __contains__(self, value):
        try:
            key = self._key(value)
        except Exception:
            return False
        return key in self._items

    def __len__(self):
        return len(self._items)

    # NOT a classmethod because self.key must be transferred too!
    # Fortunately it is always called as self._from_iterable(...)
    # in _abccoll.py
    def _from_iterable(self, iterable):
        return type(self)(iterable, key=self._key)

    def copy(self):
        return type(self)(self._items.values(), key=self._key)

    def __repr__(self):
        return "%s(%r)" % (type(self).__name__, list(self._items.values()))

    def __hash__(self):
        # Set already provides an implementation for a hash method (_hash),
        # but we have to apply it to the keys only.
        # Python 2.x requires the 'self' argument to be a Set instance,
        # so we call the underlying function directly. Python 3.x has
        # relaxed the requirement.
        _hash = Set._hash
        if hasattr(_hash, "im_func"):
            _hash = _hash.im_func
        return _hash(self._items.keys())


class KeyedSet(FrozenKeyedSet, MutableSet):
    """A mutable set that uses a custom element equality function."""

    # "named" methods like those of `set` class, in addition to operators
    intersection_update = _build_variant(MutableSet, "__iand__")
    update = _build_variant(MutableSet, "__ior__")
    difference_update = _build_variant(MutableSet, "__isub__")
    symmetric_difference_update = _build_variant(MutableSet, "__ixor__")

    __hash__ = None  # because FrozenKeyedSet implements it

    # Implementation of abstract methods from the MutableSet ABC

    def add(self, value):
        """Add an element."""
        key = self._key(value)
        if key not in self._items:
            self._items[key] = value

    def discard(self, value):
        """Remove an element.  Do not raise an exception if absent."""
        key = self._key(value)
        try:
            del self._items[key]
        except KeyError:
            pass

    # performance: override implementation in MutableSet
    def clear(self):
        self._items.clear()



import logging
import sys
from traceback import TracebackException

from .filters import AddContextualFieldFilter, MaskPasswordsInLogs
from .formatters import JSONFormatter, ConsoleFormatter
from .utils import _sanitize_stacktrace_for_json_fields


def _uncaught_exception_logger(
    type: BaseException, exc: Exception, traceback: TracebackException
):
    exception_string = _sanitize_stacktrace_for_json_fields(type, exc, traceback)
    # raise to root logger where it will be consumed by the formatter attached to our handler
    logging.getLogger().error(exception_string)


def configure_logging(
    logger_name: str, level: str, log_format: str = "json"
) -> logging.Logger:
    root_logger = logging.getLogger()
    root_logger.addHandler(logging.StreamHandler())
    root_logger.setLevel(level)

    assert root_logger.hasHandlers()
    for handler in root_logger.handlers:
        if log_format != "console":
            handler.setFormatter(JSONFormatter())
        else:
            handler.setFormatter(ConsoleFormatter())
        handler.addFilter(MaskPasswordsInLogs())
        handler.addFilter(AddContextualFieldFilter("service", logger_name))

    sys.excepthook = _uncaught_exception_logger

    return logging.getLogger(logger_name)



import calendar
from datetime import datetime, date
from enum import Enum
from functools import total_ordering


@total_ordering
class WeekDay(Enum):
    MONDAY = "MONDAY"
    TUESDAY = "TUESDAY"
    WEDNESDAY = "WEDNESDAY"
    THURSDAY = "THURSDAY"
    FRIDAY = "FRIDAY"
    SATURDAY = "SATURDAY"
    SUNDAY = "SUNDAY"

    @classmethod
    def extract_from_datetime(cls, dt: datetime):
        return cls[calendar.day_name[dt.weekday()].upper()]

    @classmethod
    def extract_from_date(cls, dt: date):
        return cls[calendar.day_name[dt.weekday()].upper()]

    @classmethod
    def from_string(cls, weekday: str):
        return cls[weekday.upper()]

    def get_day_after(self, num_of_days: int):
        index = _order.index(self)
        new_index = (index + num_of_days) % len(_order)
        return _order[new_index]

    def __str__(self):
        return str(self.name)

    def __lt__(self, other):
        return _order.index(self) < _order.index(other)

    def __sub__(self, other) -> int:
        return (_order.index(self) - _order.index(other)) % 7


_order = [
    WeekDay.MONDAY,
    WeekDay.TUESDAY,
    WeekDay.WEDNESDAY,
    WeekDay.THURSDAY,
    WeekDay.FRIDAY,
    WeekDay.SATURDAY,
    WeekDay.SUNDAY,
]



import itertools as it
import random
from typing import List, Callable, Any, Optional, Dict
from uuid import uuid4

from shuttlis.geography import Location


def one_or_none(lizt: List[Any], key: Callable[[Any], bool]) -> Optional[Any]:
    gen = (x for x in lizt if key(x))
    elm = next(gen, None)
    if elm:
        assert next(gen, None) is None
    return elm


def next_or_none(lizt: List[Any], key: Callable[[Any], bool]) -> Optional[Any]:
    gen = (x for x in lizt if key(x))
    elm = next(gen, None)
    return elm


def one(lizt: List[Any], key: Callable[[Any], bool]) -> Any:
    maybe_elm = one_or_none(lizt, key)
    assert maybe_elm is not None
    return maybe_elm


def group_by(lizt: List[Any], key: Callable[[Any], Any]) -> Dict[Any, List[Any]]:
    sorted_list = sorted(lizt, key=key)
    return {k: list(v) for k, v in it.groupby(sorted_list, key)}


def id_map(lizt: List[Any], key: Callable[[Any], Any]) -> Dict[Any, Any]:
    list_map = group_by(lizt, key=key)
    assert all(len(v) == 1 for v in list_map.values())
    return {k: v[0] for k, v in list_map.items()}


def random_location():
    lat = random.uniform(-90, 90)
    lng = random.uniform(-180, 180)
    return Location(lat, lng)


def uuid4_str():
    return str(uuid4())


def pairwise(lizt):
    "[s0, s1, s2, s3] -> (s0,s1), (s1,s2), (s2, s3), ..."  # n-1 items from list
    return [x for x in zip(lizt[:-1], lizt[1:])]
